//Authors: Dario Cattaruzza, Alessandro Abate, Peter Schrammel, Daniel Kroening
//University of Oxford 2016
//This code is supplied under the BSD license agreement (see license.txt)
#ifndef CANONICALSYSTEM_H
#define CANONICALSYSTEM_H

#include <map>
#include "DigitalSystem.h"

namespace abstract
{
template <class scalar> class CanonicalSystem : public DigitalSystem<scalar>
{
public:
    using typename JordanMatrix<scalar>::refScalar;
    using typename JordanMatrix<scalar>::type;
    using typename JordanMatrix<scalar>::func;
    using typename JordanMatrix<scalar>::complexS;
    using typename JordanMatrix<scalar>::MatrixS;
    using typename JordanMatrix<scalar>::MatrixC;
    using typename JordanMatrix<scalar>::SolverMatrixType;
    using typename JordanMatrix<scalar>::SolverComplexMatrixType;
    using typename JordanMatrix<scalar>::pseudoType_t;
    using typename AccelMatrix<scalar>::powerS;
    using typename AbstractMatrix<scalar>::powerList;
    using typename AbstractMatrix<scalar>::iteration_pair;

    using JordanMatrix<scalar>::m_dimension;
    using JordanMatrix<scalar>::m_isOne;
    using JordanMatrix<scalar>::m_hasOnes;
    using JordanMatrix<scalar>::m_hasMultiplicities;
    using JordanMatrix<scalar>::m_hasNegatives;
    using JordanMatrix<scalar>::m_jordanIndex;
    using JordanMatrix<scalar>::m_conjugatePair;
    using JordanMatrix<scalar>::m_eigenNorms;
    using JordanMatrix<scalar>::m_cosFactors;
    using AbstractMatrix<scalar>::m_eigenCloud;
    using AbstractMatrix<scalar>::m_sampleTime;
    using AbstractMatrix<scalar>::m_delayTime;
    using AbstractMatrix<scalar>::m_freq;
    using DynamicalSystem<scalar>::m_idimension;
    using DynamicalSystem<scalar>::m_fdimension;
    using DynamicalSystem<scalar>::m_odimension;
    using DynamicalSystem<scalar>::m_reachTime;
    using DynamicalSystem<scalar>::m_inputType;
    using DynamicalSystem<scalar>::m_initialState;
    using DynamicalSystem<scalar>::m_transformedInputs;
    using DynamicalSystem<scalar>::m_eigenValues;
    using DynamicalSystem<scalar>::m_eigenVectors;
    using DynamicalSystem<scalar>::m_invEigenVectors;
    using DynamicalSystem<scalar>::m_pseudoEigenValues;
    using DynamicalSystem<scalar>::m_pseudoEigenVectors;
    using DynamicalSystem<scalar>::m_invPseudoEigenVectors;
    using DynamicalSystem<scalar>::m_source;
    using DynamicalSystem<scalar>::m_name;
    using DynamicalSystem<scalar>::m_guard;
    using DynamicalSystem<scalar>::m_dynamics;
    using DynamicalSystem<scalar>::m_sensitivity;
    using DynamicalSystem<scalar>::m_inputs;
    using DynamicalSystem<scalar>::m_reference;
    using DynamicalSystem<scalar>::m_outputs;
    using DynamicalSystem<scalar>::m_outputSensitivity;
    using DynamicalSystem<scalar>::m_ioSensitivity;
    using DynamicalSystem<scalar>::m_outputGuard;
    using DynamicalSystem<scalar>::m_safeReachTube;
    using DynamicalSystem<scalar>::m_feedback;
    using DynamicalSystem<scalar>::m_numVertices;
    using DynamicalSystem<scalar>::m_accelVertices;
    using DynamicalSystem<scalar>::m_abstractVertices;
    using DynamicalSystem<scalar>::m_abstractInputVertices;
    using DynamicalSystem<scalar>::m_accelInSupports;
    using DynamicalSystem<scalar>::m_paramValues;
    using DynamicalSystem<scalar>::m_dynamicParams;
    using DynamicalSystem<scalar>::m_pAbstractReachTube;
    using DynamicalSystem<scalar>::m_maxIterations;
    using DynamicalSystem<scalar>::m_loadTime;
    using DigitalSystem<scalar>::m_eigenValueMultiplierMin;
    using DigitalSystem<scalar>::m_eigenValueMultiplierMax;

    using DynamicalSystem<scalar>::ms_logger;
    using DynamicalSystem<scalar>::ms_incremental;
    using DynamicalSystem<scalar>::ms_trace_dynamics;
    using DynamicalSystem<scalar>::ms_emptyMatrix;
    using DynamicalSystem<scalar>::ms_fullAnswers;

    using JordanMatrix<scalar>::interToRef;
    using JordanMatrix<scalar>::refToInter;
    using JordanMatrix<scalar>::makeInverse;
    using JordanMatrix<scalar>::jordanToPseudoJordan;
    using AbstractMatrix<scalar>::findIterations;
    using AbstractMatrix<scalar>::findFrequencies;
    using AbstractMatrix<scalar>::findZeniths;
    using AbstractMatrix<scalar>::addSupportsAtIteration;
    using AbstractMatrix<scalar>::addSupportsAtIterations;
    using AbstractMatrix<scalar>::isDivergent;
    using AbstractMatrix<scalar>::eigenCloudSupport;
    using AbstractMatrix<scalar>::getEigenCloud;
    using AbstractMatrix<scalar>::firstVarBlock;
    using AbstractMatrix<scalar>::binomial;
    using DynamicalSystem<scalar>::setName;
    using DynamicalSystem<scalar>::getTemplates;
    using DynamicalSystem<scalar>::getGuardPoly;
    using DynamicalSystem<scalar>::getInitPoly;
    using DynamicalSystem<scalar>::getInputPoly;
    using DynamicalSystem<scalar>::getReachPoly;
    using DynamicalSystem<scalar>::getTubePoly;
    using DynamicalSystem<scalar>::getAbsTubePoly;
    using DynamicalSystem<scalar>::getRoundedDirections;
    using DynamicalSystem<scalar>::makeLogahedralTemplates;
    using DynamicalSystem<scalar>::getAbstractReachTube;
    using DynamicalSystem<scalar>::getAbstractDynamics;
    using DynamicalSystem<scalar>::getAbstractVertices;
    using DynamicalSystem<scalar>::getAccelInSupports;
    using DynamicalSystem<scalar>::mergeAccelInSupports;
    using DynamicalSystem<scalar>::mergeAbstractSupports;
    using DynamicalSystem<scalar>::traceSupports;
    using DynamicalSystem<scalar>::getGuardedReachTube;
    using DynamicalSystem<scalar>::loadFromFile;
    using DynamicalSystem<scalar>::load;
    using DynamicalSystem<scalar>::save;
    using DynamicalSystem<scalar>::kronecker;
    using DynamicalSystem<scalar>::changeDimensions;
    using DynamicalSystem<scalar>::loadGuard;
    using DynamicalSystem<scalar>::loadDynamics;
    using DynamicalSystem<scalar>::loadInitialState;
    using DynamicalSystem<scalar>::loadSensitivities;
    using DynamicalSystem<scalar>::loadInputs;
    using DynamicalSystem<scalar>::loadSafeReachTube;
    using DynamicalSystem<scalar>::loadTemplates;
    using DynamicalSystem<scalar>::loadARMAXModel;
    using DynamicalSystem<scalar>::processError;
    using DynamicalSystem<scalar>::calculateJordanForm;
    using DynamicalSystem<scalar>::calculateMaxIterations;
    using DynamicalSystem<scalar>::calculateIterations;
    using DynamicalSystem<scalar>::setEigenSpace;
    using DynamicalSystem<scalar>::setInputType;
    using DynamicalSystem<scalar>::getAccelVertices;
    using DynamicalSystem<scalar>::getOutputReachTube;
    using DynamicalSystem<scalar>::process;
    using DigitalSystem<scalar>::getSpeed;
    using DigitalSystem<scalar>::getBits;
    using DigitalSystem<scalar>::findMaxValue;
    using DigitalSystem<scalar>::truncate;

    /// Constructs an empty buffer
    /// @param dimension dimension of the space
    CanonicalSystem(int dimension=0,int idimension=0,int odimension=0,bool trace=false);

    /// Transforms the matrix to Row Echelon Form
    /// @param matrix matrix to transform to REF
    /// @ret number of zero rows of the REF
    int toREF(MatrixS &matrix);

    /// Copies an existing object
    virtual void copy(const CanonicalSystem &source);

    /// Retrieves the characteristic polynomial coefficients of the dynamics
    MatrixS getDynamicPolynomialCoefficients(const MatrixC &eigenValues);
    MatrixS getDynamicPolynomialCoefficients();

    /// Retrieves the reachability matrix [B AB A^2B ...A^{n-1}B]
    MatrixS getReachabilityMatrix();

    /// Retrieves the reachability matrix [1 a1 a2...a_{n-1};0 1 a1 ... a_{n-2}...]
    MatrixS getCanonicalReachabilityMatrix();

    /// Retrieves the transform matrix T : z=T^{-1}x turns A,B,C into controllable canonical form
    MatrixS getReachableCanonicalTransformMatrix();

    /// Retrieves the observability matrix W=[C CA CA^2 ...CA^{n-1}]^T
    MatrixS getObservabilityMatrix();

    /// Retrieves the observability matrix [1 0 ...;a1 1 0 ...;...;a_{n-1}...a1 1]^-1
    MatrixS getInverseCanonicalObservabilityMatrix();

    /// Retrieves the transform matrix T : z=T^{-1}x turns A,B,C into observable canonical form
    MatrixS getObservableCanonicalTransformMatrix();

    /// Creates matrices T, invT , [W, invW], to make a reachable [observable] system
    void makeReachabilityMatrices(bool observer);

    /// Creates bounds for K,L, KT_c^-1 and T_oL
    void makeControlBounds(bool observer,int observerBits);

    /// Retrieves the reference gain
    MatrixS getReferenceGain();

    /// Synthesises a bound on the dynamics given a known guard and eigenvectors.
    void synthesiseDynamicBounds(AbstractPolyhedra<scalar> &result,inputType_t inputType,AbstractPolyhedra<scalar> &end);

    /// Creates the observer matrices (based on the FWL truncation)
    /// and returns the maximum error caused by the truncation
    /// for the eigenvalues of the observed system
    scalar makeObserver();

    /// Creates the observer matrices (based on the FWL truncation)
    /// and the closed loop dynamics related to that observer feedback
    /// and returns the maximum error caused by the truncation
    /// for the eigenvalues of the observed system
    scalar makeObserver(MatrixS &dynamics,MatrixS &sensitivity,bool hasNoise=false);

    /// Creates a closed dynamical system that overapproximates the system with inputs
    bool makeClosedSystem(CanonicalSystem<scalar> &closedSystem);

    /// Retrieves constraints on the controller coefficients based on the Dynamic constraints
    AbstractPolyhedra<scalar> getControllerDynBounds(AbstractPolyhedra<scalar>& reachTube,int &orBlockSize);

    /// Retrieves constraints on the controller coefficients based on the Input constraints
    AbstractPolyhedra<scalar> getControllerInBounds(AbstractPolyhedra<scalar>& reachTube);

    /// Retrieves constraints on the controller coefficients based on stability
    AbstractPolyhedra<scalar> getControllerStabilityBounds(const MatrixS &coefficients,bool nonNegative);

    /// Clips the matrix to fit the FWL given by bits
    void fitToSpec(MatrixS &matrix, int intbits,int fracbits);

    /// Limits the eigenvalues to cap and returns the corresponding polynomial
    MatrixS capEigenValues(scalar cap, bool split, bool keep);

    /// Retrives the transform matrix that renders the dynamics in reachable [observable] form
    void getCanonicalTransformMatrices(MatrixS &T, MatrixS &invT, bool observer);

    /// Retireves the largest number that need to be represented to execute the model
    refScalar findMaxValue(bool observer=true, bool states=true, bool transforms=false);

    /// Finds the eigenvalues that correspond to the observer
    /// (eigendecomposition does not return ordered eigenvalues)
    bool getObserverRows(scalar coeff,int index,const scalar &observerCoeff,refScalar &error,std::vector<int>& rows,int depth);

    /// Rescales the eigenvalues to change the granularity
    bool resample(scalar scale,MatrixS feedbackTarget,MatrixS observerTarget);

    int getTotalBits(bool observer,int &fracbits);

    /// Updates the feedback and observer matrices only if they have not been set or if they are being forced
    void updateFeedback(const MatrixS &feedback,const MatrixS &observer,bool force);
protected:
    /// Corrects the support set by the input offset
    void demergeAccelInSupports(MatrixS &supports,MatrixS &inSupports,int numTemplates);

    /// gets the smallest multiplier that minimises the eigenvalues while remaining in a safe region
    scalar getMinEigenvalueMultiplier(MatrixS &canonical,AbstractPolyhedra<scalar> &bounds);

public:
    MatrixS                     m_T;
    MatrixS                     m_W;
    MatrixS                     m_invT;     //m_T.inverse()
    MatrixS                     m_invW;     //m_W.inverse()
    MatrixS                     m_TT;       //m_T.transpose()
    MatrixS                     m_invTT;    //m_invT.transpose()
    MatrixS                     m_TinvW;    //m_T*m_invW
    MatrixS                     m_observer;
    MatrixS                     m_observerDynamics;
    MatrixS                     m_observerSensitivity;
    MatrixS                     m_observerOutputSensitivity;
    MatrixS                     m_polynomialCoefficients;
    scalar                      m_observerDynamicsError;
    scalar                      m_observerSensitivityError;
    scalar                      m_observerOutputSensitivityError;
    scalar                      m_controllerGershgorinRadius;
    scalar                      m_observerGershgorinRadius;
    EigenPolyhedra<scalar>      m_controllerBounds;
    EigenPolyhedra<scalar>      m_observerBounds;
    std::vector<int>            m_controllabilityIndices;
public:
    static canonicalType_t      ms_canonical;
};

}

#endif //CANONICALSYSTEM_H
