//Authors: Dario Cattaruzza, Alessandro Abate, Peter Schrammel, Daniel Kroening
//University of Oxford 2016
//This code is supplied under the BSD license agreement (see license.txt)
#ifndef DIGITALSYSTEM_H
#define DIGITALSYSTEM_H

#include <map>
#include "DynamicalSystem.h"

namespace abstract
{
template <class scalar> class DigitalSystem : public DynamicalSystem<scalar>
{
public:
    using typename JordanMatrix<scalar>::refScalar;
    using typename JordanMatrix<scalar>::type;
    using typename JordanMatrix<scalar>::func;
    using typename JordanMatrix<scalar>::complexS;
    using typename JordanMatrix<scalar>::MatrixS;
    using typename JordanMatrix<scalar>::MatrixC;
    using typename JordanMatrix<scalar>::SolverMatrixType;
    using typename JordanMatrix<scalar>::SolverComplexMatrixType;
    using typename JordanMatrix<scalar>::pseudoType_t;
    using typename AccelMatrix<scalar>::powerS;
    using typename AbstractMatrix<scalar>::powerList;
    using typename AbstractMatrix<scalar>::iteration_pair;

    using JordanMatrix<scalar>::m_dimension;
    using JordanMatrix<scalar>::m_hasMultiplicities;
    using JordanMatrix<scalar>::m_hasNegatives;
    using JordanMatrix<scalar>::m_jordanIndex;
    using JordanMatrix<scalar>::m_conjugatePair;
    using JordanMatrix<scalar>::m_eigenNorms;
    using JordanMatrix<scalar>::m_cosFactors;
    using AccelMatrix<scalar>::m_pseudoInvIminJ;
    using AccelMatrix<scalar>::m_invIminF;
    using AbstractMatrix<scalar>::m_eigenCloud;
    using AbstractMatrix<scalar>::m_sampleTime;
    using AbstractMatrix<scalar>::m_delayTime;
    using AbstractMatrix<scalar>::m_freq;
    using DynamicalSystem<scalar>::m_idimension;
    using DynamicalSystem<scalar>::m_fdimension;
    using DynamicalSystem<scalar>::m_odimension;
    using DynamicalSystem<scalar>::m_reachTime;
    using DynamicalSystem<scalar>::m_inputType;
    using DynamicalSystem<scalar>::m_initialState;
    using DynamicalSystem<scalar>::m_transformedInputs;
    using DynamicalSystem<scalar>::m_eigenValues;
    using DynamicalSystem<scalar>::m_eigenVectors;
    using DynamicalSystem<scalar>::m_invEigenVectors;
    using DynamicalSystem<scalar>::m_pseudoEigenValues;
    using DynamicalSystem<scalar>::m_pseudoEigenVectors;
    using DynamicalSystem<scalar>::m_invPseudoEigenVectors;
    using DynamicalSystem<scalar>::m_source;
    using DynamicalSystem<scalar>::m_name;
    using DynamicalSystem<scalar>::m_guard;
    using DynamicalSystem<scalar>::m_dynamics;
    using DynamicalSystem<scalar>::m_sensitivity;
    using DynamicalSystem<scalar>::m_inputs;
    using DynamicalSystem<scalar>::m_reference;
    using DynamicalSystem<scalar>::m_outputs;
    using DynamicalSystem<scalar>::m_outputSensitivity;
    using DynamicalSystem<scalar>::m_ioSensitivity;
    using DynamicalSystem<scalar>::m_outputGuard;
    using DynamicalSystem<scalar>::m_safeReachTube;
    using DynamicalSystem<scalar>::m_feedback;
    using DynamicalSystem<scalar>::m_numVertices;
    using DynamicalSystem<scalar>::m_accelVertices;
    using DynamicalSystem<scalar>::m_abstractVertices;
    using DynamicalSystem<scalar>::m_abstractInputVertices;
    using DynamicalSystem<scalar>::m_accelInSupports;
    using DynamicalSystem<scalar>::m_paramValues;
    using DynamicalSystem<scalar>::m_dynamicParams;
    using DynamicalSystem<scalar>::m_pAbstractReachTube;
    using DynamicalSystem<scalar>::m_maxIterations;
    using DynamicalSystem<scalar>::m_loadTime;

    using DynamicalSystem<scalar>::ms_logger;
    using DynamicalSystem<scalar>::ms_trace_dynamics;
    using DynamicalSystem<scalar>::ms_emptyMatrix;
    using DynamicalSystem<scalar>::ms_fullAnswers;

    using JordanMatrix<scalar>::interToRef;
    using JordanMatrix<scalar>::refToInter;
    using JordanMatrix<scalar>::makeInverse;
    using JordanMatrix<scalar>::calculateBlockSVD;
    using JordanMatrix<scalar>::jordanToPseudoJordan;
    using AccelMatrix<scalar>::calculateF;
    using AccelMatrix<scalar>::calculateInvIminF;
    using AccelMatrix<scalar>::calculateInvIminJ;
    using AbstractMatrix<scalar>::findIterations;
    using AbstractMatrix<scalar>::findFrequencies;
    using AbstractMatrix<scalar>::findZeniths;
    using AbstractMatrix<scalar>::addSupportsAtIteration;
    using AbstractMatrix<scalar>::addSupportsAtIterations;
    using AbstractMatrix<scalar>::isDivergent;
    using AbstractMatrix<scalar>::eigenCloudSupport;
    using AbstractMatrix<scalar>::getEigenCloud;
    using AbstractMatrix<scalar>::firstVarBlock;
    using AbstractMatrix<scalar>::binomial;
    using DynamicalSystem<scalar>::setName;
    using DynamicalSystem<scalar>::getTemplates;
    using DynamicalSystem<scalar>::getGuardPoly;
    using DynamicalSystem<scalar>::getInitPoly;
    using DynamicalSystem<scalar>::getInputPoly;
    using DynamicalSystem<scalar>::getReachPoly;
    using DynamicalSystem<scalar>::getTubePoly;
    using DynamicalSystem<scalar>::getAbsTubePoly;
    using DynamicalSystem<scalar>::getRoundedDirections;
    using DynamicalSystem<scalar>::makeLogahedralTemplates;
    using DynamicalSystem<scalar>::getAbstractReachTube;
    using DynamicalSystem<scalar>::getAbstractDynamics;
    using DynamicalSystem<scalar>::getAbstractVertices;
    using DynamicalSystem<scalar>::getAccelInSupports;
    using DynamicalSystem<scalar>::mergeAccelInSupports;
    using DynamicalSystem<scalar>::mergeAbstractSupports;
    using DynamicalSystem<scalar>::traceSupports;
    using DynamicalSystem<scalar>::getGuardedReachTube;
    using DynamicalSystem<scalar>::getDescription;
    using DynamicalSystem<scalar>::loadFromFile;
    using DynamicalSystem<scalar>::load;
    using DynamicalSystem<scalar>::save;
    using DynamicalSystem<scalar>::kronecker;
    using DynamicalSystem<scalar>::changeDimensions;
    using DynamicalSystem<scalar>::loadGuard;
    using DynamicalSystem<scalar>::loadDynamics;
    using DynamicalSystem<scalar>::loadInitialState;
    using DynamicalSystem<scalar>::loadSensitivities;
    using DynamicalSystem<scalar>::loadInputs;
    using DynamicalSystem<scalar>::loadSafeReachTube;
    using DynamicalSystem<scalar>::loadTemplates;
    using DynamicalSystem<scalar>::loadARMAXModel;
    using DynamicalSystem<scalar>::processInputs;
    using DynamicalSystem<scalar>::processError;
    using DynamicalSystem<scalar>::calculateJordanForm;
    using DynamicalSystem<scalar>::calculateMaxIterations;
    using DynamicalSystem<scalar>::calculateIterations;
    using DynamicalSystem<scalar>::setEigenSpace;
    using DynamicalSystem<scalar>::setInputType;
    using DynamicalSystem<scalar>::getAccelVertices;
    using DynamicalSystem<scalar>::getOutputReachTube;
    using DynamicalSystem<scalar>::process;

    /// Constructs an empty buffer
    /// @param dimension dimension of the space
    DigitalSystem(int dimension=0,int idimension=0,int odimension=0,bool trace=false);

    /// Copies an existing object
    virtual void copy(const DigitalSystem &source);

    /// Creates a model for the input of the closed loop
    AbstractPolyhedra<scalar> generateFeedbackInput(int fdimension=0,bool makeNoise=false,MatrixS &sensitivity=ms_emptyMatrix,scalar sampling=0,scalar delay=0);

    /// Creates a model for the quantization noise as an input specification
    AbstractPolyhedra<scalar> generateNoiseInput(scalar sampling,scalar delay);

    /// retrieves the Jordan form of the system sampled at SampleTime
    MatrixC getSampledJordanForm(scalar sampleTime);

    /// retrieves the dynamics of the system sampled at SampleTime
    MatrixS getSampledDynamics(scalar sampleTime);

    /// retrieves the delay noise matrix
    MatrixS getDelayNoise(scalar sampleTime, scalar maxDelay);

    /// turns a continuous time system into a discrete
    bool sample(scalar sampleTime,scalar delayTime=0,bool strict=true);
    bool sample(std::string &sampleTime);

    bool isSampled()       { return !func::isZero(m_sampleTime-func::ms_1); }

    /// Rescales the eigenvalues to change the granularity
    bool resample(scalar scale);

    /// turns a discrete time system into a continuous one
    void unsample(scalar sampleTime);

    /// Sets the expected maximum eigenvalue of the closed loop
    bool setSpeed(std::string &max);

    scalar getSpeed()       { return m_eigenValueCap; }

    /// Corrects the support set by the input offset
    void demergeAccelInSupports(MatrixS &supports,MatrixS &inSupports,int numTemplates);

    /// Retireves the largest number that need to be represented to execute the model
    virtual refScalar findMaxValue(bool dynamics=true, bool states=true, bool intermediate=false);

    /// Retireves the number of bits needed to represent max
    int getBits(refScalar max);

    /// Retireves the number of bits needed to represent the vertices of the polyhedra
    int getBits(AbstractPolyhedra<scalar> polyhedra);

    /// Clips the matrix to fit the FWL given by bits
    void fitToSpec(MatrixS &matrix, int totalbits,int fracbits);

    /// Normalizes the statespace
    bool scaleSystem(scalar scale=func::ms_1);

protected:
    /// Retireves the largest number that need to represent the matrix
    refScalar findMaxValue(const MatrixS &matrix,refScalar result=0,refScalar scale=1);

    /// Retireves the largest difference between any two non-zero numbers in a matrix
    scalar findRange(const MatrixS &matrix);

    /// Retireves the best scale factor usable to reach target bits
    scalar findScale(MatrixS &matrix,int targetBits, bool autoScale=false);

    /// Truncates the value to the current controller precision
    scalar truncateCoefficient(const scalar& coeff);

    /// Truncates the values of src to the current controller precision
    void truncate(MatrixS &dest,const MatrixS &src);

public:
    scalar                      m_eigenValueCap;
    scalar                      m_samplingCap;
    scalar                      m_eigenValueMultiplierMin;
    scalar                      m_eigenValueMultiplierMax;
public:
    static bvType_t             ms_bvType;
};

}

#endif //DIGITALSYSTEM_H
